fl

{
IR Temperature sensor(MLX90614ESF-BAA)
      
PropForth 5.5
2015/06/24 13:55:36

MLX90614ESF-BAA    QuickStart
           Vdd  -- 3.3V
           SCL  -- SCL
           SDA  -- SDA
           Vss  -- GND
}

\ =========================================================================== 
\ slow i2c     
\ =========================================================================== 

: err_msg ." I2C error" ;
\ If error, print message
\ ( n1 -- )   n1:t/f
: err? if err_msg cr then ;

: clr_bit_dira dira COG@ and dira COG! ;
: set_bit_dira dira COG@ or dira COG! ;

\ Start i2c-commnication 
\ This also can use SMBus device.
\ ( -- )
: _eestart
hDFFF_FFFF clr_bit_dira \ dira COG@ and dira COG!      \ Release sda
hEFFF_FFFF clr_bit_dira \ dira COG@ and dira COG!      \ Release scl
h2000_0000 set_bit_dira \ dira COG@ or dira COG!       \ Set sda to lo
h1000_0000 set_bit_dira \ dira COG@ or dira COG!       \ Set scl to lo
;

\ Stop i2c-commnication 
\ ( -- )
: _eestop
_scli                    \ Release scl 
_sdai                    \ Release sda
;

\ slow _eeread  (Hi=45usec Lo=65usec  9.09kHz)
\ ( t/f -- c1 ) flag should be true if this is the last read
: slow_eeread
_sdai 
0 8 0 do
     1 lshift _scli _sda? _sclo
     if 1 or then
loop
swap
if
     _sdai
else
     _sdao
then
_scli _sclo
;     
           
\ slow _eewrite  (Hi=40usec Lo=125usec  6.06kHz)
\ ( c1 -- t/f ) write c1 to the eeprom, true if there was an error
: slow_eewrite
h80
8 0 do
     2dup and
     if
          _sdai
     else
          _sdao
     then
     _scli _sclo 1 rshift
loop
2drop
_sdai _scli _sda? _sclo _sdao     
;

\ Write series data to register in i2c_device
\ ( n1..nn n2 n3 n4 -- t/f )   n1..nn:data  n2:number  n3:register  n4:slave_address   t/f:true if there was an error
: slow_i2c_wr_multi
\ Start I2C 
_eestart
\ Write slave address[wr], then receive Acknowledge-bit(ACK:Lo  NACK:Hi) 
_eewrite                           \ ( n1..nn n2 n3 t/f )
\ Write register                   
swap slow_eewrite or               \ ( n1..nn n2 t/f )
swap                               \ ( n1..nn t/f n2 )
\ Read n2 byte
dup 1 >
if                                 \ ( n1..nn t/f n2 )                 
     0 do                          \ ( n1..nn t/f )
          swap slow_eewrite or     \ ( n1.. nn t/f ) 
     loop
else                               \ ( n1 t/f n2 )     
     drop swap slow_eewrite or     \ ( t/f )
then               
\ Stop I2C
_eestop 
;

\ Read sereis data from series register in i2c_device
\ ( n1 n2 n3 -- n4 . . nn t/f )  n1:number  n2:register  n3:slave_address  n4..nn:series data  t/f:true if there was an error
: slow_i2c_rd_multi
\ Start I2C 
_eestart
\ Write slave address[wr], then receive Acknowledge-bit(ACK:Lo  NACK:Hi)  
 tuck slow_eewrite                 \ ( n1 n3 n2 t/f )
\ Write register
swap slow_eewrite or               \ ( n1 n3 t/f )
swap                               \ ( n1 t/f n3 )
\ Repeated Start read_process
_eestart
\ Write slave address[rd], then receive Acknowledge-bit(ACK:Lo  NACK:Hi)
 1 or 
slow_eewrite or                    \ ( n1 t/f )
\ Read (n1-1)bytes
>r                                 \ Push flag  ( n1 )
dup 1 > 
if 
     1 - 0 do 
          0 slow_eeread            \ ( n4..nn-1 )
     loop
else
     drop
then
\ Read 1byte ,then set sda to Hi(NACK:master->slave)
-1 slow_eeread                     \ ( n4..nn )
r>                                 \ Pop flag   ( n4..nn t/f )
\ Stop I2C
_eestop                                         
;


\ Display connected i2c-device's slave-address
\ This also can use SMBus device.
\ ( -- )
: i2c_detect
hex
\ Print lower
."      0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F" cr
\ ( n1 n2 ) n1:reserve address 's count n2:i2c_device's count n3:i2c_address
0 0 0 
8 0 do
     \ Print 1-line
     dup .byte h3A emit space           
     d16 0 do
          \ Start I2C 
          _eestart
          \ Write slave address[wr], then receive Acknowledge-bit(ACK:Lo  NACK:Hi)   
          dup i + 1 lshift slow_eewrite 0=                      
          if                                 \ there is ACK 
               dup i + dup .byte             \ Get slave-addres
               dup 7 > swap h78 < and        \ 0 to 7 and h78 to h7F
               if 
                    swap 1+ swap             \ Count up device (eliminate reserve number)
               else 
                    rot 1+ rot2
               then     
          else 
               ." --" 
          then 
          space 
          \ Stop I2C
          _eestop                                       
     loop
     cr
     \ next line
     d16 +                                                      
loop
drop
decimal
." i2c_device:" . cr  
0> if ." [0 - 7] and [h78 - h7F] are reserve-address" then cr
cr
;


\ ------------------ MLX90614 -------------------------

\ =========================================================================== 
\ Constants 
\ =========================================================================== 
\ Slave addres h5A for MLX90614 
hB4 wconstant MLX90614

\ eeprom register [h20 + eeprom_register]
h20 wconstant TOmax
h21 wconstant TOmin                         
h22 wconstant PWMctrl
h23 wconstant Ta_Range
h24 wconstant Ke
h25 wconstant ConfigR1
h2E wconstant SMBus_Addr
h3C wconstant ID_1
h3D wconstant ID_2
h3E wconstant ID_3
h3F wconstant ID_4

\ RAM register
3 wconstant Ambient
4 wconstant IR1
5 wconstant IR2
6 wconstant Ta
7 wconstant Tobj1
8 wconstant Tobj2

\ command
hF0 wconstant rd_flag
hFF wconstant sleep

\ Read data from RAM or eeprom in MLX90614
\ ( n1 -- n2 n3 t/f )  n1:command   n2:data[16bit] n3:PEC  t/f:true if there was an error
: MLX90614_rd
3 swap MLX90614 slow_i2c_rd_multi   \ ( LowByte HiByte n3 t/f )
err?
rot                                \ ( n3 LowByte HiByte )
rot 8 lshift or swap               \ ( n2 n3 )
;

\ Write data to RAM or eeprom in MLX90614
\ ( n1 n2 n3 -- t/f )  n1:PEC n2:data[16bit] n3:command  t/f:true if there was an error
: MLX90614_wr
>r                                 \ ( n1 n2 )
dup 8 rshift swap 3                \ ( n1 HiByte LowByte 3 )
r> MLX90614                        \ ( n1 HiByte LowByte 3 n3 SlaveAddress )
 slow_i2c_wr_multi                  \ ( t/f )
;

\ TAB
\ ( -- )
: tab 9 emit ;
: 2tab tab tab ;
\ Display value in RAM 0r eeprom register
\ ( n1 -- n2 )
: disp_reg1 MLX90614_rd drop ;
: disp_reg2 disp_reg1 . cr ;
: disp_reg3 disp_reg1 hFF and . ;

\ Message "Melexis reserved"
\ ( n1 -- )  n1:register
: mlx_msg1 dup .byte tab ." Melexis reserved:" disp_reg2 ;
: mlx_msg2 ." no        Melexis reserved" cr ;

\ Display RAM register
\ ( -- )
: disp_RAM
." Address" tab ." Contents" cr
hex
4 0 do i mlx_msg1 loop
4 .byte tab ." Raw data IR ch 1:" 4 disp_reg2
5 .byte tab ." Raw data IR ch 2:" 5 disp_reg2
6 .byte tab ." Tambient (16bit format, $27AD=10_157=-70C to $7FFF=32_767=382C):" 6 disp_reg2
7 .byte tab ." Tobject1 (16bit format, $27AD=10_157=-70C to $7FFF=32_767=382C):" 7 disp_reg2
8 .byte tab ." Tobject2 (16bit format, $27AD=10_157=-70C to $7FFF=32_767=382C):" 8 disp_reg2
9 .byte tab ." Melexis reserved:" 9 disp_reg2
hA .byte tab ." Melexis reserved, Ta1_PKI:" hA disp_reg2
hB .byte tab ." Melexis reserved, Ta2_PKI:" hB disp_reg2
h13 hC do i mlx_msg1 loop
h13 .byte tab ." Melexis reserved, Scale_Alpha_Ratio:" h13 disp_reg2
h14 .byte tab ." Melexis reserved, Scale_Alpha_Slope:" h14 disp_reg2
h15 .byte tab ." Melexis reserved, IIR_Filter:" h15 disp_reg2
h16 .byte tab ." Melexis reserved, Ta1_PKI_Fraction:" h16 disp_reg2
h17 .byte tab ." Melexis reserved, Ta2_PKI_Fraction:" h17 disp_reg2
h1B h18 do i mlx_msg1 loop
h1B .byte tab ." Melexis reserved, FIR_Filter:" h1B disp_reg2
h20 h1C do i mlx_msg1 loop
decimal
;

\ Display eeprom register
\ ( -- )
: disp_eeprom
hex
." Address" tab ." Contents" tab ." Writable" cr
0 .byte tab h20 disp_reg1 . 2tab ." yes       To_max" cr
1 .byte tab h21 disp_reg1 . 2tab ." yes       To_min" cr
2 .byte tab h22 disp_reg1 . 2tab ." yes       PWMCTRL" cr
3 .byte tab h23 disp_reg1 . 2tab ." yes       Ta range" cr
4 .byte tab h24 disp_reg1 . 2tab ." yes       emissivity correction coefficient" cr
5 .byte tab h25 disp_reg1 . 2tab ." yes       config register1" cr
h2E h26 do i h20 - .byte tab i disp_reg1 . 2tab mlx_msg2 loop
hE .byte tab h2E MLX90614_rd drop hFF and . 2tab ." yes       SMbus address (LSB only)" cr
hF .byte tab h2F disp_reg1 . 2tab ." yes       Melexis reserved" cr
h39 h30 do i h20 - .byte tab i disp_reg1 . 2tab mlx_msg2 loop
h19 .byte tab h39 disp_reg1 . 2tab ." yes       Melexis reserved" cr
h3C h3A do i h20 - .byte tab i disp_reg1 . 2tab mlx_msg2 loop
h1C .byte tab h3C disp_reg3 2tab ." no        ID number" cr
h1D .byte tab h3D disp_reg3 2tab ." no        ID number" cr
h1E .byte tab h3E disp_reg3 2tab ." no        ID number" cr
h1F .byte tab h3F disp_reg3 2tab ." no        ID number" cr
decimal
;

\ PEC uses CRC-8 to calculate PacketErrorCode(PEC).
\ ( n1 n2 -- n3 )  n1:crc n2:data  n3:crc
: crc8
xor                      \ ( result )  xor new byte and old crc to get remainder + byte
\ check all 8 bit
8 0 do
     1 lshift            \ ( result' result')  shift it out
     dup h100 and 0 <>   \ ( result' t/f )
     if
          7 xor          \ ( result'' )  xor with the polynomial
     then   
loop  
hFF and                  \ remove the bit shifted out
;     

\ Writ data in eeprom
\ ( n1 n2 -- ) n1:value[16bit] n2:eeprom address
: wr_eeprom
over dup 8 rshift swap 2 ST@                      \ ( n1 n2 Hi-byte Lo-byte n2 ) 
\ MLX90614 0 -> crc8 -> 5 n2 -> crc8 -> hD1 Lo-byte -> crc8 -> ?? Hi-byte -> crc8 -> **
MLX90614 0 crc8 crc8 crc8 crc8                    \ ( n1 n2 PEC )
rot2
MLX90614_wr                                       \ ( PEC n1 n2 )
5 delms
;  

\ Set data in eeprom
\ ( n1 n2 -- ) n1:value[16bit] n2:eeprom address
: set_eeprom
dup 0 swap wr_eeprom     \ Erase value
wr_eeprom                \ Write value
;

\ Change SA(Slave Address)
\ ( n1 -- )  n1:SA
: set_SA
dup 0 <>
if 
     0 SMBus_Addr wr_eeprom   \ Erase SA
     SMBus_Addr wr_eeprom     \ Write new SA
     if ." Success to set SA" then
else
     ." Not o"
then
cr
;

\ Display temperature
\ ( n1 -- )  n1:value
: disp_degree
d100 over >
if
     h30 emit h2E emit
     d10 u/mod h30 + emit h30 + emit
else
     d1000 over >
     if
          d100 u/mod h30 + emit h2E emit
          d10 u/mod h30 + emit h30 + emit
     else
          d10000 over >
          if
               d1000 u/mod h30 + emit 
               d100 u/mod h30 + emit h2E emit
               d10 u/mod h30 + emit h30 + emit
          else
               d10000 u/mod h30 + emit
               d1000 u/mod h30 + emit 
               d100 u/mod h30 + emit h2E emit
               d10 u/mod h30 + emit h30 + emit
          then
     then                    
then
;

\ Display temperature[degree]
\ ( n1 -- ) n1:
: calc_degree
d100 u* d50 u/ d27315 - 
dup 0 >=
if
     disp_degree
else
     h2D emit
     invert 1+
     disp_degree
then
." degreeC" 
;
     

\ Repeat displaying temperature
\ ( -- )
: demo
tab ." Raw" 2tab ." degreeC" cr
clkfreq cnt COG@ +
begin
     ." Tamb=" tab Ta MLX90614_rd drop dup . 2tab calc_degree cr
     ." Tobj1=" tab Tobj1 MLX90614_rd drop dup . 2tab calc_degree cr
     ." Tobj2=" tab Tobj2 MLX90614_rd drop dup . 2tab calc_degree cr
     cr
     clkfreq waitcnt
     fkey? swap drop
until
drop
;
